// git clone https://github.com/vectorclass/version2 vectorclass
// g++ -Ivectorclass -Wall -Wpedantic -Wextra -std=c++17 -march=haswell -O3 vectorclass_demo_asm.S vectorclass_demo.cpp && ./a.out

#include <random>
#include <algorithm>
#include <iostream>
#include <vector>
#include <numeric>
#include <cassert>
#include <cstdint>
#include <ctime>
#include "vectorclass.h"

using namespace std;
#define LENGTH 12800000
#define REPS 100

extern "C" { float array_sum_asm(float* array, uint64_t length); }

float array_sum_loop(float* array, uint64_t length) {
    float total = 0;
    for (size_t i = 0; i < length; i++) {
        total += array[i];
    }
    return total;
}

float array_sum_vc(float* array, uint64_t length) {
    assert(length % 8 == 0);
    Vec8f acc = Vec8f(0);
    Vec8f tmp;
    for (uint64_t i = 0; i < length; i += 8) {
        tmp.load(array + i);
        acc += tmp;
    }
    return horizontal_add(acc);
}

void time_array_sum(string name, float array_sum(float*, uint64_t), float* arr, uint64_t len) {
    struct timespec start, end;
    float result = 0;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    for ( int i=0; i<REPS; i++ ) {
        result += array_sum(arr, len);
    }
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    double elapsed = (end.tv_sec - start.tv_sec) + (end.tv_nsec - start.tv_nsec) / 1e9;
    printf("%s %g in %g ns/elt\n", name.c_str(), result, elapsed/(len*REPS)*1e9);
}

void print_float_vec(Vec8f v) {
    cout << "{ ";
    for( unsigned i=0; i < v.size(); i++ ) {
        cout << v[i] << ' ';
    }
    cout << "}\n";
}

void print_double_vec(Vec4d v) {
    cout << "{ ";
    for( unsigned i=0; i < v.size(); i++ ) {
        cout << v[i] << ' ';
    }
    cout << "}\n";
}

void examples() {
    Vec8f acc = Vec8f(0.0);
    Vec8f numbers = Vec8f(0, 1, 2, 3, 4, 5, 6, 7);
    print_float_vec(acc);
    acc += numbers;
    acc = acc * numbers;
    print_float_vec(acc);
    
    //Vec8f singles = Vec8f(0.0);
    //Vec4d doubles = Vec4d(0.0);
    //doubles += singles;
    
    double array[8] = {0, 10, 20, 30, 40, 50, 60, 70};
    Vec4d a, b;
    a.load(array);
    b.load(array + 4);
    print_double_vec(a);
    print_double_vec(b);

}

int main(void) {
    examples();
    
    // https://stackoverflow.com/a/23143753
    random_device rnd_device;
    mt19937 mersenne_engine {rnd_device()};
    uniform_real_distribution<float> dist {-0.25, 0.75};

    auto gen = [&dist, &mersenne_engine]() {
        return dist(mersenne_engine);
    };

    unsigned length = LENGTH;
    vector<float> vec(length);
    cout << "Array size: " << sizeof(float) * length / 1024 / 1024 << "MB\n";
    generate(begin(vec), end(vec), gen);

    float *array = &vec[0];
    time_array_sum("array_sum_loop", array_sum_loop, array, length);
    time_array_sum("array_sum_vc", array_sum_vc, array, length);
    time_array_sum("array_sum_asm", array_sum_asm, array, length);
    return 0;
}
